@@ -1,5 +1,6 @@
 /*
- * Copyright (c) 2009-2014 Kazuho Oku, Tokuhiro Matsuno, Daisuke Murase
+ * Copyright (c) 2009-2014 Kazuho Oku, Tokuhiro Matsuno, Daisuke Murase,
+ *                         Shigeo Mitsunari
  *
  * The software is licensed under either the MIT License (below) or the Perl
  * license.
@@ -24,9 +25,12 @@
  */
 
 #include <stddef.h>
+#ifdef __SSE4_2__
+# include <x86intrin.h>
+#endif
 #include "picohttpparser.h"
 
-/* $Id: ee9904f6a48d2bfe3d8773fe2618bd7b533028e0 $ */
+/* $Id: 4f5c43608c95934446dbfd147b115b406384b5ed $ */
 
 #if __GNUC__ >= 3
 # define likely(x) __builtin_expect(!!(x), 1)
@@ -53,8 +57,13 @@
 
 #define ADVANCE_TOKEN(tok, toklen) do { \
     const char* tok_start = buf; \
-    for (; ; ++buf) { \
+    static const char ranges2[] __attribute__((aligned(16))) = "\000\040\177\177"; \
+    int found2; \
+    buf = findchar_fast(buf, buf_end, ranges2, sizeof(ranges2) - 1, &found2); \
+    if (! found2) { \
       CHECK_EOF(); \
+    } \
+    while (1) { \
       if (*buf == ' ') { \
         break; \
       } else if (unlikely(! IS_PRINTABLE_ASCII(*buf))) { \
@@ -63,6 +72,8 @@
           return NULL; \
         } \
       } \
+      ++buf; \
+      CHECK_EOF(); \
     } \
     tok = tok_start; \
     toklen = buf - tok_start; \
@@ -78,12 +89,50 @@ static const char* token_char_map =
   "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"
   "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
 
+static const char* findchar_fast(const char* buf, const char* buf_end, const char *ranges, size_t ranges_size, int* found)
+{
+  *found = 0;
+#if __SSE4_2__
+  if (likely(buf_end - buf >= 16)) {
+    __m128i ranges16 = _mm_loadu_si128((const __m128i*)ranges);
+
+    size_t left = (buf_end - buf) & ~15;
+    do {
+      __m128i b16 = _mm_loadu_si128((void*)buf);
+      int r = _mm_cmpestri(ranges16, ranges_size, b16, 16, _SIDD_LEAST_SIGNIFICANT | _SIDD_CMP_RANGES | _SIDD_UBYTE_OPS);
+      if (unlikely(r != 16)) {
+        buf += r;
+        *found = 1;
+        break;
+      }
+      buf += 16;
+      left -= 16;
+    } while (likely(left != 0));
+  }
+#endif
+  return buf;
+}
+
 static const char* get_token_to_eol(const char* buf, const char* buf_end,
                                     const char** token, size_t* token_len,
                                     int* ret)
 {
   const char* token_start = buf;
   
+#ifdef __SSE4_2__
+  static const char ranges1[] =
+    "\0\010"
+    /* allow HT */
+    "\012\037"
+    /* allow SP and up to but not including DEL */
+    "\177\177"
+    /* allow chars w. MSB set */
+    ;
+  int found;
+  buf = findchar_fast(buf, buf_end, ranges1, sizeof(ranges1) - 1, &found);
+  if (found)
+    goto FOUND_CTL;
+#else
   /* find non-printable char within the next 8 bytes, this is the hottest code; manually inlined */
   while (likely(buf_end - buf >= 8)) {
 #define DOIT() if (unlikely(! IS_PRINTABLE_ASCII(*buf))) goto NonPrintable; ++buf
@@ -97,6 +146,7 @@ static const char* get_token_to_eol(const char* buf, const char* buf_end,
     }
     ++buf;
   }
+#endif
   for (; ; ++buf) {
     CHECK_EOF();
     if (unlikely(! IS_PRINTABLE_ASCII(*buf))) {
@@ -211,14 +261,21 @@ static const char* parse_headers(const char* buf, const char* buf_end,
       /* parsing name, but do not discard SP before colon, see
        * http://www.mozilla.org/security/announce/2006/mfsa2006-33.html */
       headers[*num_headers].name = buf;
-      for (; ; ++buf) {
+      static const char ranges1[] __attribute__((aligned(16))) = "::\x00\037";
+      int found;
+      buf = findchar_fast(buf, buf_end, ranges1, sizeof(ranges1) - 1, &found);
+      if (! found) {
         CHECK_EOF();
+      }
+      while (1) {
         if (*buf == ':') {
           break;
         } else if (*buf < ' ') {
           *ret = -1;
           return NULL;
         }
+        ++buf;
+        CHECK_EOF();
       }
       headers[*num_headers].name_len = buf - headers[*num_headers].name;
       ++buf;
